@@ -1,5 +1,5 @@
 # vi:fdm=marker fdl=0
-# $Id: SGF2misc.pm,v 1.46 2004/03/23 14:29:59 jettero Exp $ 
+# $Id: SGF2misc.pm,v 1.71 2004/03/25 17:35:52 jettero Exp $ 
 
 package Games::Go::SGF2misc;
 
@@ -8,9 +8,11 @@ no warnings;
 
 use Carp;
 use Parse::Lex;
+use Data::Dumper;
+use Compress::Zlib;
 
-# This is actually my major version, followed by my current CVS revision.
-our $VERSION = q($Revision: 1.46 $); $VERSION =~ s/[^\.\d]//g; $VERSION =~ s/^1\./0.6./;
+# This is actually my major and minor versions, followed by my current CVS revision.
+our $VERSION = q($Revision: 1.71 $); $VERSION =~ s/[^\.\d]//g; $VERSION =~ s/^1\./0.7./;
 
 1;
 
@@ -23,6 +25,7 @@ sub new {
     if( $ENV{DEBUG} > 0 ) {
         use Number::Format;
         use Devel::Size qw(total_size);
+        use Time::HiRes qw(time);
 
         $this->{frm} = new Number::Format;
     }
@@ -35,30 +38,13 @@ sub parse {
     my $this = shift;
     my $file = shift;
 
-    my %did = ();
-    my @to_nuke;
-    
-    push @to_nuke, (@{$this->{gametree}}) if ref($this->{gametree}) eq "ARRAY";
-    push @to_nuke, $this->{parse}         if ref($this->{parse})    eq "HASH";
-
-    while( @to_nuke ) {
-        my $ref = shift @to_nuke;
-        for my $k (qw(p c kids parent)) {
-            if( my $v = $ref->{$k} ) {
-                if( ref($v) eq "ARRAY" ) {
-                    push @to_nuke, @$v;
-                }
-
-                delete $ref->{$k};
-            }
-        }
-    }
-
     for my $k (keys %$this) {
-        delete $this->{$k} unless {frm=>1}->{$k};
+        delete $this->{$k} unless {Time=>1, frm=>1}->{$k};
     }
     $global::lex_error = undef;
 
+    $this->_time("parse");
+
     if( -f $file ) {
         my @rules = (
             VALUE  => '\[(?ms:.*?(?<!\x5c))\]',
@@ -153,16 +139,50 @@ sub parse {
 
         close SGFIN;
 
+        $this->_time("parse");
+
         print STDERR "SGF Parsed!  Calling internal _parse() routine\n" if $ENV{DEBUG} > 0;
         print STDERR "\$this size (before _parse)= ", $this->{frm}->format_bytes(total_size( $this )), "\n" if $ENV{DEBUG} > 0;
 
+        $this->_time("_parse");
+
         my $r = $this->_parse(0, $this->{parse});
 
+        $this->_time("_parse");
+
         print STDERR "\$this size (after _parse)= ", $this->{frm}->format_bytes(total_size( $this )), "\n" if $ENV{DEBUG} > 0;
 
         print STDERR "rebuilding {refdb} (for ref2id/id2ref)\n" if $ENV{DEBUG} > 0;
-        $this->{nodelist} = { map {$this->_ref2id($_) => $this->_node_list([], $_)} @{$this->{gametree}} };
-        print STDERR "\$this size (after _node_list())= ", $this->{frm}->format_bytes(total_size( $this )), "\n" if $ENV{DEBUG} > 0;
+
+        $this->_time("_nodelist");
+
+        $this->{nodelist} = { map {$this->_ref2id($_) => $this->_nodelist([], $_)} @{$this->{gametree}} };
+
+        $this->_time("_nodelist");
+
+        print STDERR "\$this size (after _nodelist())= ", $this->{frm}->format_bytes(total_size( $this )), "\n" if $ENV{DEBUG} > 0;
+
+        $this->_time("nuke(gametree and parse)");
+        my @to_nuke;
+        
+        push @to_nuke, (@{$this->{gametree}}) if ref($this->{gametree}) eq "ARRAY";
+        push @to_nuke, $this->{parse}         if ref($this->{parse})    eq "HASH";
+
+        while( @to_nuke ) {
+            my $ref = shift @to_nuke;
+            for my $k (qw(p c kids parent)) {
+                if( my $v = $ref->{$k} ) {
+                    if( ref($v) eq "ARRAY" ) {
+                        push @to_nuke, @$v;
+                    }
+
+                    delete $ref->{$k};
+                }
+            }
+        }
+        $this->_time("nuke(gametree and parse)");
+
+        $this->_show_timings if $ENV{DEBUG} > 0;
 
         return $r;
     }
@@ -171,6 +191,69 @@ sub parse {
     return 0;
 }
 # }}}
+# freeze {{{
+sub freeze {
+    my $this = shift;
+
+    local $Data::Dumper::Indent = 0;
+    local $Data::Dumper::Purity = 1;
+
+    my $fm = {};
+    for my $k (qw(nodelist refdb)) {
+        $fm->{$k} = $this->{$k};
+    }
+
+    $this->_time("freeze Dumper");
+    my $buf = Dumper( $fm );
+    $this->_time("freeze Dumper");
+
+    return Compress::Zlib::memGzip( $buf );
+}
+# }}}
+# thaw {{{
+sub thaw {
+    my $this = shift;
+    my $frz  = shift;
+    my ($VAR1);
+
+    if( ref($frz) eq "GLOB" ) {
+        $this->_time("gzreads");
+        my $gz = gzopen($frz, "r"); 
+
+        $frz = "";
+
+        my $b;
+        while( my $r = $gz->gzread($b, 32768) ) {
+            $frz .= $b;
+        }
+        $gz->gzclose;
+
+        $this->_time("gzreads");
+
+        $this->_time("eval");
+        eval $frz;
+        $this->_time("eval");
+    } else {
+        $this->_time("memgunzip/eval");
+        eval Compress::Zlib::memGunzip( $frz );
+        $this->_time("memgunzip/eval");
+        if( $@ ) {
+            $this->{error} = $@;
+            return 0;
+        }
+    }
+
+    $this->_time("assign refs");
+    for my $k (keys %$VAR1) {
+        $this->{$k} = $VAR1->{$k};
+    }
+    $this->_time("assign refs");
+
+    $this->_show_timings if $ENV{DEBUG} > 0;
+
+    return 1;
+}
+# }}}
 # errstr {{{
 sub errstr {
     my $this = shift;
@@ -237,28 +320,37 @@ sub parse_hash {
     return $this->{parse};
 }
 # }}}
-# node_list {{{
-sub node_list {
+# nodelist {{{
+sub nodelist {
     my $this  = shift;
 
     return $this->{nodelist};
 }
 # }}}
+# is_node {{{
+sub is_node {
+    my $this = shift;
+    my $node = shift;
+
+    return ($this->{refdb}{$node} ? 1:0);
+}
+# }}}
 # as_perl {{{
 sub as_perl {
     my $this = shift;
     my $node = shift;
+    my $soft = shift;
 
     if( $node ) {
-        if( my $ref = $this->_id2ref( $node ) ) {
+        if( my $ref = $this->{refdb}{$node} ) {
             return $ref;
-        } else {
-            $this->{error} = "Error: no such node($node).";
-            return 0;
         }
     }
 
-    return (wantarray ? @{$this->{gametree}} : $this->{gametree});
+    $this->{error} = "no such node: $node";
+    return 0 if $soft;
+
+    croak $this->{error};
 }
 # }}}
 # as_text {{{
@@ -266,12 +358,7 @@ sub as_text {
     my $this = shift;
     my $node = shift;
 
-    if( $node =~ m/\d+\-(\d+|root)/ ) {
-       $node = $this->as_perl( $node ) or return 0;
-    } elsif( not ref($node) eq "HASH" ) {
-        $this->{error} = "The as_functions() take only take IDs or node references as arguments, ie, not \"$node\"";
-        return 0;
-    }
+    $node = $this->as_perl( $node, 1 ) or croak $this->errstr;
 
     my $board = $node->{board};
 
@@ -293,12 +380,7 @@ sub as_html {
     my $dir  = shift;
        $dir  = "./img" unless $dir;
 
-    if( $node =~ m/\d+\-(\d+|root)/ ) {
-       $node = $this->as_perl( $node ) or return 0;
-    } elsif( not ref($node) eq "HASH" ) {
-        $this->{error} = "The as_functions() take only take IDs or node references as arguments";
-        return 0;
-    }
+    $node = $this->as_perl( $node, 1 ) or croak $this->errstr;
 
     my $board = $node->{board};
     my $size  = @{$board->[0]}; # inaccurate?
@@ -378,15 +460,11 @@ sub as_html {
 # as_image {{{
 sub as_image {
     my $this = shift;
-    my $node = shift;
+    my $node = shift; my $nm = $node;
     my $argu = shift;
     my %opts = (imagesize=>256, antialias=>0);
 
-    if( $node =~ m/\d+\-(\d+|root)/ ) {
-       $node = $this->as_perl( $node ) or return 0;
-    } elsif( not ref($node) eq "HASH" ) {
-        croak "The as_functions() take only take IDs or node references as arguments";
-    }
+    $node = $this->as_perl( $node, 1 ) or croak $this->errstr;
 
     my $board = $node->{board};
     my $size  = @{$board->[0]}; # inaccurate?
@@ -394,27 +472,12 @@ sub as_image {
     if( ref($argu) eq "HASH" ) {
         @opts{keys %$argu} = (values %$argu);
     } else {
-        croak "as_image() takes a hashref argument... e.g., {imagesize=>256, etc=>1} or nothing at all.";
+        croak 
+        "as_image() takes a hashref argument... e.g., {imagesize=>256, etc=>1} or nothing at all.";
     }
 
     $opts{boardsize} = $size;
-    $opts{filename}  = "$node->{variation}-$node->{move_no}.png" unless $opts{filename};
-
-    my $name = "Games::Go::SGF2misc::GD (aka as_image)";
-
-    eval "use GD;"; 
-    if( $@ ) {
-        croak "You need to install GD/GD::Image (version 2.15+ or so) to use $name";
-    } elsif( $GD::VERSION < 2.0 ) {
-        croak "You need GD/GD::Image version 2.15+ or so to use $name";
-    }
-
-    eval "use Games::Go::SGF2misc::GD;";
-    if( $@ ) {
-        if( $@ =~ m/Bareword/ ) {
-            croak "You need GD/GD::Image version 2.15+ or so to use $name";
-        } else { croak $@ }
-    }
+    $opts{filename}  = "$nm.png" unless $opts{filename};
 
     use Games::Go::SGF2misc::GD;
     my $image = new Games::Go::SGF2misc::GD(%opts);
@@ -428,7 +491,9 @@ sub as_image {
                 if( $ENV{DEBUG} > 0 ) {
                     print STDERR "placeStone($1, [$i, $j])\n";
                 }
-                $image->placeStone(lc($1), [reverse( $i, $j )]);  # SGFs are $y, $x, my matrix is $x, $y ...
+
+                # SGFs are $y, $x, the matrix is $x, $y ...
+                $image->placeStone(lc($1), [reverse( $i, $j )]);  
             }
         }
     }
@@ -436,34 +501,116 @@ sub as_image {
     # draw moves
     # draw marks
 
+    if( $opts{filename} =~ m/^\-\.(\w+)$/ ) {
+        return $image->dump($1);
+    }
+
     $image->save($opts{filename});
 }
 # }}}
+# as_freezerbag {{{
+sub as_freezerbag {
+    my $this = shift;
+    my $file = shift or croak "You must name your freezerbag.";
+    my $code = shift;
+       $code = "# your code here\n" unless $code;
+    my $perl = shift;
+
+    if( not $perl ) {
+        for my $try (qw{ /usr/bin/perl /usr/local/bin/perl }) {
+            $perl = $try if -x $try;
+        }
+        croak "couldn't find perl" unless -x $perl;
+    }
+    
+    open  OUTMF, ">$file" or croak "Couldn't open freezerbag ($file) for output: $!";
+    print OUTMF "#!$perl\n# vi:fdm=marker fdl=0:\n\nuse strict;\nno warnings;\nuse Games::Go::SGF2misc;\n\n";
+    print OUTMF "my \$sgf = new Games::Go::SGF2misc;\n";
+    print OUTMF "   \$sgf->thaw(\\*DATA);\n\n$code\n\n# freezer DATA {\{\{\n__DATA__\n";
+
+    $this->_time("print freeze");
+    print OUTMF $this->freeze;
+    $this->_time("print freeze");
+
+    close OUTMF;
+
+    $this->_show_timings if $ENV{DEBUG} > 0;
+}
+# }}}
 
 # internals
-# _node_list {{{
-sub _node_list {
+# _show_timings {{{
+sub _show_timings {
+    my $this = shift;
+
+    my @times = ();
+    for my $k (keys %{ $this->{Time} }) {
+        my $a   = $this->{Time}{$k}{diffs};  next unless ref($a) eq "ARRAY";
+        my $n   = int @$a;
+        my $sum = 0;
+           $sum += $_ for @$a;
+
+        push @times, [ $k, $sum, $n, ($sum/$n) ];
+    }
+
+    for my $a (sort {$b->[1] <=> $a->[1]} @times) {
+        printf('%-35s: sum=%3.4fs cardinality=%5d avg=%3.2fs%s', @$a, "\n");
+    }
+
+    delete $this->{Time};
+}
+# }}}
+# _time {{{
+sub _time {
+    return unless $ENV{DEBUG} > 0;
+
+    my $this = shift; 
+    my $tag  = shift;
+
+    if( $ENV{DEBUG} == 1.2 ) {
+        my @a;
+
+        for (sort keys %{ $this->{Time} }) {
+           push @a, $_ if $this->{Time}{$_}{start};
+        }
+
+        print STDERR "clocks: @a\n";
+    }
+
+    if( defined $this->{Time}{$tag}{start} ) {
+        push @{ $this->{Time}{$tag}{diffs} }, (time - $this->{Time}{$tag}{start});
+        delete $this->{Time}{$tag}{start};
+    } else {
+        $this->{Time}{$tag}{start} = time;
+    }
+}
+# }}}
+# _nodelist {{{
+sub _nodelist {
     my $this = shift;
     my $list = shift;
     my $cur  = shift;
 
-    # $this->{nodelist} = { map {$this->_ref2id($_) => $this->_node_list([], $_)} @{$this->{gametree}} };
+    # $this->{nodelist} = { map {$this->_ref2id($_) => $this->_nodelist([], $_)} @{$this->{gametree}} };
 
     for my $kid (@{ $cur->{kids} }) {
         my $id = $this->_ref2id( $kid );
-        my ($var, $mov) = split /\-/, $id;
 
-        if( $var > @{ $list } ) {
+        die "problem parsing node id" unless $id =~ m/(\d+)\.(\d+)\-(.+)/;
+
+        my ($g, $v, $m) = ($1, $2, $3);
+
+        if( $v > @{ $list } ) {
             my $b = [];
             push @$list, $b;
-            for (1..$mov) {
+            for (1..$m) {
                 push @$b, undef;
             }
         }
 
-        push @{ $list->[$var-1] }, $id;
+        push @{ $list->[$v-1] }, $id;
 
-        $this->_node_list($list, $kid);
+        $this->_nodelist($list, $kid);
     }
 
     return $list;
@@ -508,6 +655,7 @@ sub _parse {
         # Then we re _parse() at our current position
 
         $gref = { variations=>1, kids=>[] }; push @{ $this->{gametree} }, $gref;
+        $gref->{gnum} = int @{ $this->{gametree} };
 
         my $pnode = $pref->{n}[0];
         for my $p (@$pnode) {
@@ -556,14 +704,16 @@ sub _parse {
         for my $i (0..$#{ $pref->{n} }) {
             my $pnode = $pref->{n}[$i];
 
-            $gnode = { variation=>$gref->{variations}, parent => ($gnode ? $gnode : $parent ? $parent : $gref), kids=>[] };
-            push @{ $gnode->{parent}{kids} }, $gnode;
+            $parent = ($gnode ? $gnode : $parent ? $parent : $gref);
 
-            $gnode->{board} = $this->_copy_board_matrix( $gnode->{parent}{board} ) if $gnode->{parent}{board};
+            $gnode = { variation=>$gref->{variations}, kids=>[] };
+            push @{ $parent->{kids} }, $gnode;
+
+            $gnode->{board} = $this->_copy_board_matrix( $parent->{board} ) if $parent->{board};
             $gnode->{board} = $this->_new_board_matrix( $gref ) unless $gnode->{board};
 
             $gnode->{captures} = { B=>0, W=>0 };
-            if( ref($gnode->{parent}) and ref(my $pc = $gnode->{parent}{captures}) ) {
+            if( ref($parent) and ref(my $pc = $parent->{captures}) ) {
                 $gnode->{captures}{B} += $pc->{B};
                 $gnode->{captures}{W} += $pc->{W};
             }
@@ -619,9 +769,10 @@ sub _parse {
                 }
             }
 
+            $gnode->{gnum}    = $parent->{gnum};
             $gnode->{move_no} = 
                   (ref($gnode->{moves}) ? int(@{ $gnode->{moves} }) : 0)
-                + (ref($gnode->{parent}) and defined $gnode->{parent}{move_no} ? $gnode->{parent}{move_no} : 0);
+                + (ref($parent) and defined $parent->{move_no} ? $parent->{move_no} : 0);
         }
 
         my $j = @{ $pref->{c} };
@@ -648,46 +799,47 @@ sub _parse {
     return 0;
 }
 # }}}
-# _id2ref {{{
-sub _id2ref {
-    my $this = shift;
-    my $id   = shift;
-
-    return $this->{refdb}{$id} if defined $this->{refdb}{$id};
-    return undef;
-}
-# }}}
 # _ref2id {{{
 sub _ref2id {
     my $this = shift;
     my $ref  = shift;
 
-    unless( defined $this->{refdb}{$ref} ) {
+    croak "invalid ref given to _ref2id()" unless ref($ref) eq "HASH";
+
+    unless( defined $this->{refdb2}{$ref} ) {
         my $id;
         my $c = 2;
         if( defined($ref->{variation}) and defined($ref->{move_no}) ) {
-            $id = $ref->{variation} . "-" . ($ref->{move_no} ? $ref->{move_no} : "root");
+            $id = "$ref->{gnum}." . 
+                   $ref->{variation} . "-" . ($ref->{move_no} ? $ref->{move_no} : "root");
             my $cur = $id;
             while( defined $this->{refdb}{$cur} ) {
                 $cur = $id . "-" . $c++;
             }
             $id = $cur;
         } else {
-            $id = "game #" . (++$this->{games});
+            $id = ++$this->{games};
         }
 
         print STDERR "$ref 2 id: $id\n" if $ENV{DEBUG} >= 10;
 
-        $this->{refdb}{$ref} = $id;
-        $this->{refdb}{$id} = $ref;
+        $this->{refdb2}{$ref} = $id;
+
+        for my $k (qw(board marks moves other captures game_properties variations)) {
+            $this->{refdb}{$id}{$k} = $ref->{$k} if defined $ref->{$k};
+        }
+
+        for my $k (qw(gnum kids)) {
+            delete $this->{refdb}{$id}{$k};
+        }
 
         if( $ENV{DEBUG} > 20 ) {
-            print STDERR "\$this\->\{refdb\}\{\$ref($ref)\} = $this->{refdb}{$ref} ",
+            print STDERR "\$this\->\{refdb\}\{\$ref($ref)\} = $this->{refdb2}{$ref} ",
                         "/ \$this\-\>\{refdb\}\{\$id($id)\} = $this->{refdb}{$id}\n";
         }
     }
 
-    return $this->{refdb}{$ref};
+    return $this->{refdb2}{$ref};
 }
 # }}}
 # _new_board_matrix {{{
@@ -695,6 +847,8 @@ sub _new_board_matrix {
     my $this = shift;
     my $gref = shift;
 
+    $this->_time("_new_board_matrix");
+
     my $board = [];
 
     my $size = $gref->{game_properties}{SZ};
@@ -708,6 +862,8 @@ sub _new_board_matrix {
         push @$board, $row;
     }
 
+    $this->_time("_new_board_matrix");
+
     return $board;
 }
 # }}}
@@ -716,27 +872,19 @@ sub _copy_board_matrix {
     my $this = shift;
     my $tocp = shift;
 
+    $this->_time("_copy_board_matrix");
+
     my $board = [];
 
-    for my $i (0..$#{ $tocp }) {
-        my $row = [];
-        for my $j (0..$#{ $tocp->[$i] }) {
-            push @$row, $tocp->[$i][$j];
-        }
-        push @$board, $row;
+    my $double_check = int @$tocp;
+    for (@$tocp) {
+        my @a = @{ $_ }; 
+        push @$board, \@a;
+
+        die "Problem copying board!" unless int @a == $double_check;
     }
 
-    my $max_i = $#{ $tocp };
-    for my $i (0 .. $max_i) {
-        if( $max_i != $#{ $tocp->[$i] } ) {
-            if( $ENV{DEBUG} > 0 ) {
-                use Data::Dumper;
-                $Data::Dumper::Indent = 1;
-                print STDERR Dumper( $board ), "\n";
-            }
-            die "FATAL: problem copying board";
-        }
-    }
+    $this->_time("_copy_board_matrix");
 
     return $board;
 }
@@ -748,6 +896,8 @@ sub _check_for_captures {
     my $board = $node->{board};
     my $caps  = $node->{captures};
 
+    $this->_time("_check_for_captures");
+
     my $tc = $board->[$p[0]][$p[1]];
 
     croak "crazy unexpected error: checking for caps, and current pos doesn't have a stone.  Two times double odd, and fatal" 
@@ -757,6 +907,8 @@ sub _check_for_captures {
 
     # 1. Find groups for all adjacent stones.  
 
+    $this->_time("for(_find_group)");
+
     my %checked = ();
     my @groups  = ();
     for my $p ( [$p[0]-1, $p[1]+0], [$p[0]+1, $p[1]+0], [$p[0]+0, $p[1]-1], [$p[0]+0, $p[1]+1] ) {
@@ -765,6 +917,9 @@ sub _check_for_captures {
         push @groups, [ @g ] if @g;
     }
 
+    $this->_time("for(_find_group)");
+    $this->_time("for(\@groups), _count_liberties");
+
     if( @groups ) {
         # 2. Any groups without liberties are toast!
         print STDERR "_check_for_captures() found ", int(@groups), " neighboring groups:" if $ENV{DEBUG} > 3 and int(@groups);
@@ -785,6 +940,9 @@ sub _check_for_captures {
         print STDERR "\n" if $ENV{DEBUG} > 3;
     }
 
+    $this->_time("for(\@groups), _count_liberties");
+    $this->_time("_find_group/_count_liberties of me");
+
     # 3. Check my own liberties, I may be toast
     %checked = ();
     my @me_group = $this->_find_group( \%checked, $SZ, $tc, $board, @p );
@@ -799,12 +957,17 @@ sub _check_for_captures {
         }
     }
     print STDERR "\n" if $ENV{DEBUG}>3;
+
+    $this->_time("_find_group/_count_liberties of me");
+    $this->_time("_check_for_captures");
 }
 # }}}
 # _count_liberties {{{
 sub _count_liberties {
     my ($this, $SZ, $board, @group) = @_;
 
+    $this->_time("_count_liberties");
+
     my %checked = ();
     my $count   = 0;
 
@@ -821,6 +984,8 @@ sub _count_liberties {
         }
     }
 
+    $this->_time("_count_liberties");
+
     return $count;
 }
 # }}}
@@ -828,6 +993,8 @@ sub _count_liberties {
 sub _find_group {
     my ($this, $checked, $SZ, $oc, $board, @p) = @_;
 
+    $this->_time("_find_group");
+
     print STDERR "\t_find_group(@p)" if $ENV{DEBUG}>12;
     my @g;
 
@@ -847,6 +1014,8 @@ sub _find_group {
     }
     print STDERR "\n" if $ENV{DEBUG}>12;
 
+    $this->_time("_find_group");
+
     return @g;
 }
 # }}}
@@ -870,136 +1039,42 @@ __END__
         # the output options are listed below
     }
 
-=head1 as_perl
+=head1 nodelist
 
-    my @gametree = $sgf->as_perl; # smart enough to return an arrayref or array as appropriate
-    my $gametree = $sgf->as_perl; # the arrayref is probably a better choice, since it's not a copy.
+    my $nodelist = $sgf->nodelist;  
 
-    # This will probably be the most useful as_ function (if you're doing
-    # your own custom thing).  It returns all the nodes, beautifully
-    # parsed, with the board positions as a big matrix, kibitzen as an
-    # array of strings, the list of moves, the list of captures, and the
-    # rules, etc.
-    # 
-    # Could you ask for anything more?  
-    # 
-    # Less memory usage perhaps?  Because of the huge amount of
-    # semi-duplicate data in the tree, these take up (very roughly) around
-    # a meg of ram per 300 move game.
+    # This returns a special list of node-id's (human readable
+    # $game_number.$variation-$move_no node descriptors).  example:
 
-=head2 The Layout (Pretty Much)
-
-    # The game tree is actually an array(ref) of games.
-
-    $gametree = [ $game1, $game2, $game3, $game4 ];
-
-    # The games are hashes (refs) of game data with the Collections of
-    # Nodes stuffed into a kids=>[] array(ref).
-
-    $game = { variations=>6, 
-        game_properties={
-          'AP' => 'CGoban:2', 'FF' => 3, 'PB' => 'Myself', 'GM' => 1, 'KM' => '0', 'SZ' => 5,
-          'RU' => 'Japanese', 'CA' => 'UTF-8', 'PW' => 'Me'
-        },
-        kids => [ $node1, $node2, $node3 ],
-    };
-
-    # And, lastly, the nodes are hashes with their own kids array.
-    $node = {
-        parent=>$game, # this is actually a pointer (ref) to the hashref
-                       # who's kids=>[] array points to this node.
-
-        'variation' => 1,
-
-        'other' => { 'ST' => '2' },  # cgoban 1 and 2 kick this out.  *shrug*
-                                     # the other properties are just values
-                                     # SGF2misc doesn't handle
-
-        'move_no' => 1, # The root node is move #0. 
-                        # It is technically possible to have more than one
-                        # move in a given node.  If so, the move counter is
-                        # bumped up to count the last move of the node.
-
-        'kids' => [ $node1, $node2, $etc ],
-
-        # This should be pretty clear.  SGF2misc leaves the sgf
-        # co-ordinates intact, but also presents some numerical ones for
-        # your enjoyment.  CR is a circle, btw.  The numerical co-ordinates
-        # are for the board matrix and start in the upper left corner
-        # (unlike the letter-number co-ordinates we use IRL).
-
-        'marks' => [ [ 'CR', 1, 3, 'bd' ] ],
-        'moves' => [ [ 'B', 1, 3, 'bd' ] ],
-
-        # This is from the root node of a handicap game... They're the handicap stones.
-        'edits' => [ [ 'B', 15, 3, 'pd' ], [ 'B', 3, 15, 'dp' ] ],
-
-        # The board is a matrix of descriptive characters.  This is a 3x3
-        # board, and clearly, the edits from above do not fit on it.
-        'board' => [ [' ', ' ', ' '], [' ', ' ', ' '], [' ', ' ', ' '] ], 
-
-        # This speaks for itself I hope.
-        'captures' => { B=>3, W=>4 },
-
-        # I generally like to say something friendly at the start of all my games, you?
-        'comments' => [ 'jettero [15k]: hi ' ],
-    }
-
-=head2 Lemme See It
-
-    # Rather than explain the layout in detail, it's better to simply see
-    # it.  WARNING:  This can be gigantically huge!  In memory, a complete
-    # game is probably around a meg.  When you print them though, the
-    # indenting get's out of control!
-
-    use Data::Dumper;
-    $Data::Dumper::Maxdepth = 10; # set this to 0 to see the whole show...
-    $Data::Dumper::Indent   = 1;  
-    $Data::Dumper::Purity   = 1 if $Data::Dumper::Maxdepth == 0;
-
-    # Purity makes some of the crazy-refs show up at the end instead
-    # crazy refs?  EG: 'parent' =>
-    # $VAR1->[0]{'kids'}[0]{'kids'}[0]{'kids'}[0]{'kids'}[0]{'kids'}[0]{'kids'}[0]{'kids'}[0],
-
-    open OUT, ">darray.pl" or die $!;
-    print OUT Dumper( $gametree );
-    clsoe OUT;
-
-    # One last thing.  You can check the memory usage by using the DEBUG
-    # environment variable (see below).  Anything over a 1 will show how
-    # big the structures get at both parse phasen.
-
-=head1 node_list
-
-    my $node_list = $sgf->node_list;  
-
-    # Parsing the game tree from as_perl() could be pretty tedius.
-    # Fortunately, you don't have to.  This returns a special list
-    # of node-id's (human readable $variation-$move_no node descriptors).
-    # example:
-
-    $node_list = { 'game #1' => [
-        [ '1-root', '1-1', '1-2', '1-3', '1-4', '1-5', ],
-        [    undef, undef, undef, undef, undef, '2-5', '2-6', '2-7', ],
-        [    undef, undef, undef, undef, undef, undef, undef, '3-7', '3-8' ],
+    $nodelist = { 1 => [
+        [ '1.1-root', '1.1-1', '1.1-2', '1.1-3', '1.1-4', '1.1-5', ],
+        [      undef,   undef,   undef,   undef,   undef, '1.2-5', '1.2-6', '1.2-7', ],
+        [      undef,   undef,   undef,   undef,   undef,   undef,   undef, '1.3-7', '1.3-8' ],
     ] };
 
-    # As confusing as you found the as_perl() above, you'll probably like this.
-    # as_perl() understands these node identifiers!
+    # What do you do with them?
 
-    my $game_info = $sgf->as_perl('game #1');
-    my $root_node = $sgf->as_perl('1-root');
-    my $move5_v2  = $sgf->as_perl('2-5');
+=head1 as_perl
+
+    my $game_info = $sgf->as_perl(1);
+    my $root_node = $sgf->as_perl('1.1-root');
+    my $move5_v2  = $sgf->as_perl('1.2-5');
 
     my $s = $game_info->{game_properties}{SZ};
     print "The board size is: ${s}x${s}!\n"; 
 
     print my $c (@{ $move5_v2->{comments} }) {
-        print "Comment from 2-5: $c\n";
+        print "Comment from 1.2-5: $c\n";
     }
 
     # Tada!!
 
+    # Oh, as_perl takes an optional second argument.
+    
+    $sgf->as_perl("doesn't exist node"); # will die on the spot...
+    $sgf->as_perl("isn't there", 1)      # will not
+        or print "yikes!!: " . $sgf->errstr;
+
 =head1 as_text 
 
     # This is pretty much just an example
@@ -1010,8 +1085,8 @@ __END__
     my $sgf = new Games::Go::SGF2misc; 
        $sgf->parse("sgf/jettero-sixrusses-2004-03-18.sgf");
 
-    my $nl  = $sgf->node_list;
-    my $end = $nl->{'game #1'}[0][ $#{$nl->{'game #1'}[0]} ];
+    my $nl  = $sgf->nodelist;
+    my $end = $nl->{1}[0][ $#{$nl->{1}[0]} ];
               # 1st game  1st variation    last node
      
     my $caps = $sgf->as_perl( $end )->{captures};
@@ -1049,8 +1124,9 @@ __END__
     print OUT, $sgf->as_html( $end, "/image/dir/" );
 
     # The only real difference is the image-dir argument (which defaults to
-    # "./img").  There is a directory of images included with SGF2misc.
-    # They are from this page:
+    # "./img").  This is the URL image dir (<img # src="$image_dir/moku.gif">), 
+    # not the pysical image dir.  There is a directory of images included
+    # with SGF2misc.  They are from this page:
 
     # http://www.britgo.org/gopcres/gopcres1.html#ag-gifs
 
@@ -1070,66 +1146,65 @@ __END__
 =head1 as_image
 
     # This uses the fantastic ::SGF2misc::GD package by Orien Vandenbergh
-    # that comes with this package.
+    # that comes with this package.  It is a separate package, and as you
+    # will see, the interfaces aren't totally compatable -- however, it
+    # _is_ intended to be used with this package.
 
-    # You must install GD-2.15 (or so) in order to use it!!
-    # You will need the latest bleeding edge versions of libpng and libgd
-    # though.  At the time of this writing, I used libgd-2.0.12 and got
-    # GD-2.15 to install and function normally.
+    # You must install GD-2.15 (or so) in order to use it!!  You will also
+    # need the bleeding edge versions of libpng and libgd.  At the time of
+    # this writing, I used libgd-2.0.12 and got GD-2.15 to install and
+    # function normally.
 
-    # Rather than explain myself like a normal human, I'll provide a
-    # lengthy example:
+    # Here's a calling example:
+    $sgf->as_image($node, {filename=>"html/$a->[$i].png", gobanColor=>[255, 255, 255]});
 
-    my $sgf = new Games::Go::SGF2misc;
-       $sgf->parse($ARGV[0]) or warn "could not parse: " . $sgf->errstr;
+    # ::SGF2misc::GD takes hash-like arguments.  So, so does as_image().
+    # filename=>"" and gobanColor=>[] are additions of mine, as they're
+    # actually used on separate calls in the ::SGF2misc::GD package.
+    # All other arguments are passed to the new() member function if
+    # ::SGF2misc::GD.  Please read the Games::Go::SGF2misc::GD manpage; it
+    # contains much more information.
 
+    # Also, in the interests of making this as clumsy as possible, if the
+    # filename is a dash followed by a type extension,
 
-    # But I'll explain a copule things.  This fetches the nodelist for the
-    # first variation in the SGF.  This is not necessarilly the longest
-    # variation, but it is the first variation.
+        my $image = $sgf->as_image($l, {filename=>"-.png"});
 
-    my $a = $sgf->node_list->{"game #1"}->[0];
+    # then the image will be returned as a string rather than written to a
+    # file.  
 
-    for my $i (0..$#{$a}) {
-        open OUT, ">html/$a->[$i].html" or die $!;
+=head1 as_freezerbag
+  
+   # What?  Yeah, this is a special way to save the parsed sgf data such
+   # that you can interact with it without having to re-parse it.  It
+   # seemed like it should be faster than re-parsing the SGF every time,
+   # but it isn't much of a speedup at all.  Unless I can speed it up, or
+   # someone asks me to leave it in, it could get axed.  I may leave it in
+   # anyway, because it's cute.
 
-        if( $i-1 >= 0 ) {
-            print OUT "<li><a href=\"", $a->[$i-1], ".html\">previous</a>";
-        } else { print OUT "<li> previous" }
+   # Try it yourself:
 
-        if( $i+1 <= $#{$a} ) {
-            print OUT "<li><a href=\"", $a->[$i+1], ".html\">next</a>"
-        } else { print OUT "<li> next" }
+    $sgf->as_freezerbag( "freezer.pl",  # an export filename
 
+        # Some code to put in it.  This argument is optional.
+            q/my @nodes = @{ $sgf->nodelist->{1}[0] };
 
-        # ::SGF2misc::GD takes hash-like arguments.  So, so does
-        # as_image().  It must write to a file, and doesn't return anything
-        # useful.
+            for my $n (@nodes) {
+                my $board = $sgf->as_text($n) or die $sgf->errstr; 
+                print "\e[1;1f$board";
+            }/,
 
-        # filename=>"" and gobanColor=>[] are additions of mine, as they're
-        # actually used on separate calls in the ::SGF2misc::GD package.
-        # All other arguments are passed to the new() member function.
+       # The location and/or switches for your perl
+       # I needed blib/lib (since I don't always install to test this
+       # stuff).  This argument is optional.
 
-        # For further information, please read the Games::Go::SGF2misc::GD
-        # manpage.
-
-        $sgf->as_image($a->[$i], {filename=>"html/$a->[$i].png", gobanColor=>[255, 255, 255]});
-
-        print OUT "<P>";
-        print OUT "<img src=\"$a->[$i].png\">";
-
-        close OUT;
-    }
-
-=head1 parse_hash
-
-    my $hash = $sgf->parse_hash;  
-
-    # You'll find this highly useless.  It returns the parse tree as a perl
-    # hash.  Check out as_perl() instead.
+       "/usr/bin/perl -I blib/lib",
+    );
 
 =head1 Board Postion Character Map
 
+    # This is how an empty 3x3 board would be stored:
+
     $board = [
         [ ' ', ' ', ' ' ],
         [ ' ', ' ', ' ' ],
@@ -1143,6 +1218,12 @@ __END__
     # Marks are not placed on the board!
     # You'll just have to fetch the marks array from the $node.
 
+=head1 Miscellaneous
+
+=head2 is_node
+
+    print "There is a node called 1.1-root!\n" if $sgf->is_node('1.1-root');
+
 =head1 BUGS
 
     Besides the lack of documentation?  Well, I'm sure there's a bunch.
